"""
# Dimensionality transformations

This module contains classes for dimensionality transformations, such as encoders and decoders.
These are to be used in the context where reduced dynamics is sought after,
but only microscopic data is available.

The main classes are `Encoder` and `Decoder`, which are abstract classes that should be
implemented by the user. The `ClosureEncoder` and `ClosureDecoder` classes are concrete
implementations of the `Encoder` and `Decoder` classes, respectively.
They are used to encode and decode the data into a reduced space,
after which the reduced dynamics can be trained.
In this case the reduced space consist of both known macroscopic coordinates
and learned closure coordinates.

The `Encoder` and `Decoder` objects are used in `onsagernet.dynamics.ReducedSDE`
to be used for closure modelling, for example.

```python
from onsagernet.transformations import ClosureEncoder, ClosureDecoder
from onsagernet.dynamics = ReducedSDE, OnsagerNet

encoder = ClosureEncoder(...)
decoder = ClosureDecoder(...)
sde = OnsagerNet(...)

reduced_sde = ReducedSDE(encoder, decoder, sde)  # can be used to train, predict, etc
```

As in `onsagernet.dynamics`, only the model assembly logic is provided here.
Some example implementations of the model architecture are provided
in the `onsagernet.models` module.

"""

import equinox as eqx
import jax.numpy as jnp
from abc import abstractmethod

# ------------------------- Typing imports ------------------------- #
from jax import Array
from jax.typing import ArrayLike

# ------------------------------------------------------------------ #
#                        Encoders and decoders                       #
# ------------------------------------------------------------------ #


class Encoder(eqx.Module):
    """The base class for encoders."""

    @abstractmethod
    def __call__(self, x: ArrayLike) -> Array:
        pass


class Decoder(eqx.Module):
    """The base class for decoders."""

    @abstractmethod
    def __call__(self, z: ArrayLike) -> Array:
        pass


class EncoderfromFunc(Encoder):
    """Encoder constructed from a given closure transform."""

    closure_transform: eqx.Module

    def __init__(self, closure_transform: eqx.Module) -> None:
        r"""Encoder constructed from a given closure transform.

        Takes a given closure transformation $x\mapsto z$
        to define the encoder.

        Args:
            closure_transform (eqx.Module): a given transformation $x\mapsto z$
        """
        self.closure_transform = closure_transform

    def __call__(self, x: ArrayLike) -> Array:
        return self.closure_transform(x)


class DecoderfromFunc(Decoder):
    """Decoder constructed from a given inverse closure transform."""

    inverse_closure_transform: eqx.Module

    def __init__(self, inverse_closure_transform: eqx.Module) -> None:
        r"""Decoder constructed from a given inverse closure transform.

        Takes a given inverse closure transformation $z\mapsto x$

        Args:
            inverse_closure_transform (eqx.Module): a given transformation $z\mapsto x$
        """
        self.inverse_closure_transform = inverse_closure_transform

    def __call__(self, z: ArrayLike) -> Array:
        return self.inverse_closure_transform(z)


class ClosureEncoder(EncoderfromFunc):
    """Closure encoder which combines known macroscopic coordinates
    with learned (or PCA) closure coordinates.
    """

    macroscopic_transform: eqx.Module

    def __init__(
        self, macroscopic_transform: eqx.Module, closure_transform: eqx.Module
    ) -> None:
        r"""Closure encoder which combines known macroscopic coordinates
        with learned (or PCA) closure coordinates.

        $$
            x \mapsto z = [\varphi^*(x), \hat\varphi(x)]
        $$

        where $\varphi^*$ is the known macroscopic transformation and
        $\hat\varphi$ is the learned closure transformation.

        Args:
            macroscopic_transform (eqx.Module): the known macroscopic transformation
            closure_transform (eqx.Module): the learned closure transformation
        """
        self.macroscopic_transform = macroscopic_transform
        self.closure_transform = closure_transform

    def __call__(self, x: ArrayLike) -> Array:
        """Combines the macroscopic and closure coordinates.

        Args:
            x (ArrayLike): miroscopic state

        Returns:
            Array: reduced state
        """
        macroscopic_coords = self.macroscopic_transform(x)
        closure_coords = self.closure_transform(x)
        reduced_coords = jnp.concatenate([macroscopic_coords, closure_coords])
        return reduced_coords


class ClosureDecoder(DecoderfromFunc):
    """Decodes from a closure encoder model output."""

    macroscopic_dim: int

    def __init__(
        self, inverse_closure_transform: eqx.Module, macroscopic_dim: int
    ) -> None:
        r"""Closure decoder which extracts the closure coordinates from the reduced state
        and then applies the inverse closure transformation to reconstruct the microscopic state.

        $$
            z[\text{macroscopic_dim}:] \mapsto x
        $$

        It is assuemd that the first `macroscopic_dim` coordinates are the known
        macroscopic coordinates and the rest are the learned closure coordinates.

        Args:
            inverse_closure_transform (eqx.Module): transformation from closure coordinates to microscopic state
            macroscopic_dim (int): the dimension of the known macroscopic state
        """
        self.inverse_closure_transform = inverse_closure_transform
        self.macroscopic_dim = macroscopic_dim

    def __call__(self, z: ArrayLike) -> Array:
        """Extracts the closure coordinates and applies the inverse closure transformation.

        Args:
            z (ArrayLike): reduced coordinates

        Returns:
            Array: reconstructed microscopic state
        """
        z_closure = z[self.macroscopic_dim :]
        return self.inverse_closure_transform(z_closure)



"""
# Custom loss functions

This module contains custom loss functions for training the models.

The case class `Loss` is an abstract class that defines the interface for loss functions.
Sub-classes must implement the `Loss.compute_sample_loss` method.

# """

import jax
import equinox as eqx
import jax.numpy as jnp
from abc import abstractmethod
from jax.scipy.stats import multivariate_normal
from jax.typing import ArrayLike
from jax import Array
from .dynamics import SDE, ReducedSDE


class Loss(eqx.Module):
    """Base class for losses

    Computes the loss over a batch of samples.pdoc --math my_module

    Args:
        model (eqx.Module): the model to be trained
        data_arrs (tuple[ArrayLike, ...]): input data arrays of the format (t, x, args)

    Returns:
        float: loss value
    """

    @abstractmethod
    def compute_sample_loss(
        self, model: eqx.Module, *processed_data_arrs: tuple[ArrayLike, ...]
    ) -> Array:
        """Computes the loss for a single sample.

        Args:
            model (eqx.Module): the model to be trained
            processed_data_arrs (tuple[ArrayLike, ...]): processed data arrays

        Returns:
            Array: the loss value
        """
        pass

    def _process_data_arrs(
        self, *data_arrs: tuple[ArrayLike, ...]
    ) -> tuple[ArrayLike, ...]:
        """Process the data arrays to a format that can be used by the loss function.

        The input data arrays are in the standard format of (t, x, args).
        The format of the output arrays depend on the loss.

        Args:
            data_arrs (tuple[ArrayLike, ...]): input data arrays

        Returns:
            tuple[ArrayLike, ...]: processed data arrays
        """
        return data_arrs

    @eqx.filter_jit
    def __call__(self, model: eqx.Module, *data_arrs: tuple[ArrayLike, ...]) -> float:
        data_arrs_processed = self._process_data_arrs(*data_arrs)
        in_axes = (None,) + len(data_arrs_processed) * (0,)
        compute_loss_over_time = jax.vmap(self.compute_sample_loss, in_axes=in_axes)
        compute_loss = jax.vmap(compute_loss_over_time, in_axes=in_axes)
        sample_losses = compute_loss(model, *data_arrs_processed)

        return jnp.mean(sample_losses)


class MLELoss(Loss):
    r"""Computes the maximum likelihood estimation loss for the SDE

    The loss is the negative log-likelihood of the data given the model.
    By an Euler-Maruyama discretization of the SDE
    $$
        dX(t) = f(t, X(t), \theta) dt + g(t, X(t), \theta) dW(t),
    $$
    which gives
    $$
        X(t + \Delta t) = X(t) + f(t, X(t), \theta) \Delta t + g(t, X(t), \theta) \Delta W(t),
    $$
    where $\Delta W(t) \sim N(0, \Delta t)$.
    Thus, the negative log-likelihood $X(t+\Delta t)$ is given by
    $$
        -\log p(X(t + \Delta t) | X(t), \theta) = \frac{1}{2} \log(2\pi)
        + \frac{1}{2} \log(\text{det}(\Sigma))
        + \frac{1}{2} (X(t + \Delta t) - X(t))^T \Sigma^{-1} (X(t + \Delta t) - X(t)).
    $$
    We will use `scipy.stats.multivariate_normal` to compute the log-likelihood.
    """

    def _process_data_arrs(self, t: ArrayLike, x: ArrayLike, args: ArrayLike) -> Array:
        return t[:, :-1, :], t[:, 1:, :], x[:, :-1, :], x[:, 1:, :], args[:, :-1, :]

    def compute_sample_loss(
        self,
        model: SDE,
        t: ArrayLike,
        t_plus: ArrayLike,
        x: ArrayLike,
        x_plus: ArrayLike,
        args: ArrayLike,
    ) -> float:
        """Computes the loss for a single sample.

        Args:
            model (SDE): the model to be trained
            t (ArrayLike): time
            t_plus (ArrayLike): time shifted by one step
            x (ArrayLike): state
            x_plus (ArrayLike): state shifted by one step
            args (ArrayLike): arguments for the current time step

        Returns:
            float: loss value
        """
        drift = model.drift(t, x, args)
        diffusion = model.diffusion(t, x, args)
        dt = t_plus - t
        data = (x_plus - x) / dt
        mean = drift
        cov = (1 / dt) * diffusion @ diffusion.T
        return -multivariate_normal.logpdf(data, mean, cov)


class ReconLoss(Loss):
    r"""Computes the reconstruction loss for the encoder and decoder components.

    The loss is defined by
    $$
        \| X - \text{decoder}(\text{encoder}(X)) \|^2.
    $$
    """

    def compute_sample_loss(self, model: ReducedSDE, x: ArrayLike) -> float:
        """Computes the loss for a single sample.

        Args:
            model (ReducedSDE): the model to be trained
            x (ArrayLike): state

        Returns:
            float: loss value
        """
        z = model.encoder(x)
        x_recon = model.decoder(z)
        return jnp.mean((x - x_recon) ** 2)


class CompareLoss(Loss):
    r"""Computes the comparison loss for the encoder and decoder components against PCA.

    This loss is defined by
    $$
        \max(0, \log(\text{recon_loss_model}) - \log(\text{recon_loss_pca})),
    $$
    where `recon_loss_model` is the reconstruction loss for the model
    and `recon_loss_pca` is the reconstruction loss for PCA.
    """

    def compute_sample_loss(self, model: ReducedSDE, x: ArrayLike) -> float:
        """Computes the loss for a single sample.

        Args:
            model (ReducedSDE): the model to be trained
            x (ArrayLike): state

        Returns:
            float: loss value
        """
        x_recon_model = model.decoder(model.encoder(x))
        x_recon_pca = model.decoder.inverse_closure_transform.inverse_pca_transform(
            model.encoder.closure_transform.pca_transform(x)
        )

        recon_loss_model = jnp.mean((x - x_recon_model) ** 2)
        recon_loss_pca = jnp.mean((x - x_recon_pca) ** 2)

        return jax.nn.relu(jnp.log(recon_loss_model) - jnp.log(recon_loss_pca))



"""
# Custom equinox modules for OnsagerNet components

This module contains custom equinox modules for the components of the OnsagerNet model,
which are used in various examples provided in the repository.

For new applications, it is suggested to try the simple models here first
and then build upon them, adapting if necessary to the specific problem at hand.





"""

import jax
import jax.numpy as jnp
import equinox as eqx

from ._activations import get_activation
from ._layers import ConstantLayer

# ------------------------- Typing imports ------------------------- #

from jax import Array
from jax.typing import ArrayLike
from typing import Callable
from jax.random import PRNGKey


# ------------------------------------------------------------------ #
#                           Template models                          #
# ------------------------------------------------------------------ #


class MLP(eqx.Module):
    """Multi-layer perceptron."""

    layers: list[eqx.nn.Linear]
    activation: Callable[[ArrayLike], Array]

    def __init__(
        self, key: PRNGKey, dim: int, units: list[int], activation: str
    ) -> None:
        r"""Multi-layer perceptron.

        Example:
        `mlp = MLP(key=jax.random.PRNGKey(0), dim=2, units=[32, 32, 1], activation='tanh')`
        gives a 2-hidden-layer MLP

        $$2 \to 32 \to 32 \to 1$$

        with tanh activation.

        Args:
            key (PRNGKey): random key
            dim (int): dimension of the input
            units (list[int]): layer sizes
            activation (str): activation function (can be any in `jax.nn` or custom ones defined in `onsagernet._activations`)
        """
        num_layers = len(units)
        units = [dim] + units
        keys = jax.random.split(key, num_layers)
        self.layers = [
            eqx.nn.Linear(units[i], units[i + 1], key=key) for i, key in enumerate(keys)
        ]
        self.activation = get_activation(activation)

    def __call__(self, x: ArrayLike) -> Array:
        h = x
        for layer in self.layers[:-1]:
            h = self.activation(layer(h))
        output = self.layers[-1](h)
        return output


# ------------------------------------------------------------------ #
#                         Potential networks                         #
# ------------------------------------------------------------------ #


class PotentialMLP(MLP):
    """Potential network based on a multi-layer perceptron."""

    alpha: float
    dim: int
    param_dim: int

    def __init__(
        self,
        key: PRNGKey,
        dim: int,
        units: list[int],
        activation: str,
        alpha: float,
        param_dim: int = 0,
    ) -> None:
        r"""Potential network based on a multi-layer perceptron.

        This implements the potential function
        $$
            V(x, args) = \alpha \|(x, args)\|^2 + \text{MLP}(x, args)
        $$
        where $x$ is the input and $u$ are additional parameters.
        The constant $\alpha \geq 0$ is a regularisation term,
        which gives a quadratic growth to ensure that the potential is integrable.
        We are tacitly assuming that MLP is of sub-quadratic growth,
        so only choose activation functions that have this property
        (most activation functions are either bounded or of linear growth).

        Args:
            key (PRNGKey): random key
            dim (int): dimension of the input
            units (list[int]): layer sizes
            activation (str): activation function (can be any in `jax.nn` or custom ones defined in `onsagernet._activations`)
            alpha (float): regulariser
            param_dim (int, optional): dimensions of the parameters. Defaults to 0.
        """
        self.dim = dim
        units = units + [1]
        self.param_dim = param_dim
        super().__init__(key, dim + param_dim, units, activation)
        self.alpha = alpha

    def __call__(self, x: ArrayLike, args: ArrayLike) -> Array:
        if self.param_dim > 0:
            x = jnp.concatenate([x, args[1:]], axis=0)
        output = super().__call__(x) + self.alpha * (x @ x)
        return jnp.squeeze(output)


class PotentialResMLP(MLP):
    r"""Potential network with a residual connection."""

    alpha: float
    gamma_layer: eqx.nn.Linear
    dim: int
    param_dim: int

    def __init__(
        self,
        key: PRNGKey,
        dim: int,
        units: list[int],
        activation: str,
        n_pot: int,
        alpha: float,
        param_dim: int = 0,
    ) -> None:
        r"""Potential network with a residual connection.

        This implements the modified potential function
        $$
            V(x, args) = \alpha \|(x, args)\|^2
            + \frac{1}{2}
            \| \text{MLP}(x, args)+ \Gamma (x, args) \|^2
        $$
        where

        - $\phi$ is a MLP of dim + param_dim -> n_pot
        - $\Gamma$ ia matrix of size [n_pot, dim + para_dim]
        - $\alpha > 0$ is a scalar regulariser

        Args:
            key (PRNGKey): random key
            dim (int): dimension of the input
            units (list[int]): layer sizes
            activation (str): activation function (can be any in `jax.nn` or custom ones defined in `onsagernet._activations`)
            n_pot (int): size of the MLP part of the potential
            alpha (float): regulariser
            param_dim (int, optional): dimension of the parameters. Defaults to 0.
        """
        self.dim = dim
        self.param_dim = param_dim
        units = units + [n_pot]
        mlp_key, gamma_key = jax.random.split(key)
        super().__init__(mlp_key, dim + param_dim, units, activation)
        self.alpha = alpha
        self.gamma_layer = eqx.nn.Linear(
            dim + param_dim, n_pot, key=gamma_key, use_bias=False
        )

    def __call__(self, x: ArrayLike, args: ArrayLike) -> Array:
        if self.param_dim > 0:
            x = jnp.concatenate([x, args[1:]], axis=0)
        output_phi = super().__call__(x)
        output_gamma = self.gamma_layer(x)
        output_combined = (output_phi + output_gamma) @ (output_phi + output_gamma)
        regularisation = self.alpha * (x @ x)
        return 0.5 * output_combined + regularisation


# ------------------------------------------------------------------ #
#                        Dissipation networks                        #
# ------------------------------------------------------------------ #


class DissipationMatrixMLP(MLP):
    """Dissipation matrix network based on a multi-layer perceptron."""

    alpha: float
    is_bounded: bool
    dim: int

    def __init__(
        self,
        key: PRNGKey,
        dim: int,
        units: list[int],
        activation: str,
        alpha: float,
        is_bounded: bool = True,
    ) -> None:
        r"""Dissipation matrix network based on a multi-layer perceptron.

        The MLP maps $x$ of dimension `dim` to a matrix $L(x)$ of size `dim` x `dim`,
        and then reshapes it to a `dim` x `dim` matrix.
        Then, the output matrix is given by
        $$
            M(x) = \alpha I + L(x) L(x)^\top.
        $$
        If `is_bounded` is set to `True`, then the output is element-wise bounded
        by applying a `jax.nn.tanh` activation to the output matrix $L$.

        Args:
            key (PRNGKey): random key
            dim (int): dimension of the input
            units (list[int]): layer sizes
            activation (str): activation function (can be any in `jax.nn` or custom ones defined in `onsagernet._activations`)
            alpha (float): regulariser
            is_bounded (bool, optional): whether to give a element-wise bounded output. Defaults to True.
        """
        self.dim = dim
        units = units + [dim * dim]
        super().__init__(key, dim, units, activation)
        self.alpha = alpha
        self.is_bounded = is_bounded

    def __call__(self, x: ArrayLike) -> Array:
        L = super().__call__(x).reshape(self.dim, self.dim)
        if self.is_bounded:
            L = jax.nn.tanh(L)
        return self.alpha * jnp.eye(self.dim) + L @ L.T


# ------------------------------------------------------------------ #
#                        Conservation networks                       #
# ------------------------------------------------------------------ #


class ConservationMatrixMLP(MLP):
    """Conservation matrix network based on a multi-layer perceptron."""

    is_bounded: bool
    dim: int

    def __init__(
        self,
        key: PRNGKey,
        dim: int,
        activation: str,
        units: list[int],
        is_bounded: bool = True,
    ) -> None:
        r"""Conservation matrix network based on a multi-layer perceptron.

        The MLP maps $x$ of dimension `dim` to a matrix $L(x)$ of size `dim` x `dim`,
        and then reshapes it to a `dim` x `dim` matrix.
        Then, the output matrix is given by
        $$
            W(x) = L(x) - L(x)^\top.
        $$
        If `is_bounded` is set to `True`, then the output is element-wise bounded
        by applying a `jax.nn.tanh` activation to the output matrix $L$.

        Args:
            key (PRNGKey): random key
            dim (int): dimension of the input
            activation (str): activation function (can be any in `jax.nn` or custom ones defined in `onsagernet._activations`)
            units (list[int]): layer sizes
            is_bounded (bool, optional): whether to give a element-wise bounded output. Defaults to True.
        """
        self.dim = dim
        units = units + [dim * dim]
        super().__init__(key, dim, units, activation)
        self.is_bounded = is_bounded

    def __call__(self, x: ArrayLike) -> Array:
        L = super().__call__(x).reshape(self.dim, self.dim)
        if self.is_bounded:
            L = jax.nn.tanh(L)
        return L - L.T


# ------------------------------------------------------------------ #
#                         Diffusion networks                         #
# ------------------------------------------------------------------ #


class DiffusionMLP(MLP):
    """Diffusion matrix network based on a multi-layer perceptron."""

    alpha: float
    dim: int
    param_dim: int

    def __init__(
        self,
        key: PRNGKey,
        dim: int,
        units: list[int],
        activation: str,
        alpha: float,
        param_dim: int = 0,
    ) -> None:
        r"""Diffusion matrix network based on a multi-layer perceptron.

        This implements the diffusion matrix function
        $$
            \sigma(x, args) = \text{Chol}(\alpha I + \text{MLP}(x, args))
        $$
        where $\text{Chol}$ is the Cholesky decomposition.
        Here, MLP maps $(x, args)$ of dimension `dim` + `param_dim` to a matrix of size `dim` x `dim`,

        Args:
            key (PRNGKey): random key
            dim (int): dimension of the input
            units (list[int]): layer sizes
            activation (str): activation function (can be any in `jax.nn` or custom ones defined in `onsagernet._activations`)
            alpha (float): regulariser
            param_dim (int, optional): dimension of the parameters. Defaults to 0.
        """
        self.dim = dim
        self.param_dim = param_dim
        units = units + [dim * dim]
        super().__init__(key, dim + param_dim, units, activation)
        self.alpha = alpha

    def __call__(self, x: ArrayLike, args: ArrayLike) -> Array:
        if self.param_dim > 0:
            x = jnp.concatenate([x, args[1:]], axis=0)
        sigma = super().__call__(x).reshape(self.dim, self.dim)
        sigma_squared_regularised = self.alpha * jnp.eye(self.dim) + sigma @ sigma.T
        return jnp.linalg.cholesky(sigma_squared_regularised)


class DiffusionDiagonalMLP(MLP):
    """Diagonal diffusion matrix network based on a multi-layer perceptron."""

    alpha: float
    dim: int
    param_dim: int

    def __init__(
        self,
        key: PRNGKey,
        dim: int,
        units: list[int],
        activation: str,
        alpha: float,
        param_dim: int = 0,
    ) -> None:
        r"""Diagonal diffusion matrix network based on a multi-layer perceptron.

        This implements the diffusion matrix function
        $$
            \sigma(x, args) = \text{diag}(\alpha + \text{MLP}(x, args)^2)^{\frac{1}{2}}.
        $$
        Here, MLP maps $(x, args)$ of dimension `dim` + `param_dim` to a vector of size `dim`.

        Args:
            key (PRNGKey): random key
            dim (int): dimension of the input
            units (list[int]): layer sizes
            activation (str): activation function (can be any in `jax.nn` or custom ones defined in `onsagernet._activations`)
            alpha (float): regulariser
            param_dim (int, optional): dimension of the parameters. Defaults to 0.
        """
        self.dim = dim
        self.param_dim = param_dim
        units = units + [dim]
        super().__init__(key, dim + param_dim, units, activation)
        self.alpha = alpha

    def __call__(self, x: ArrayLike, args: ArrayLike) -> Array:
        if self.param_dim > 0:
            x = jnp.concatenate([x, args[1:]], axis=0)
        sigma_diag = super().__call__(x)
        sigma_diag_regularised = jnp.sqrt(self.alpha + sigma_diag**2)
        return jnp.diag(sigma_diag_regularised)


class DiffusionDiagonalConstant(eqx.Module):
    """Diagonal diffusion matrix network based on a constant layer."""

    alpha: float
    constant_layer: ConstantLayer
    dim: int

    def __init__(self, key: PRNGKey, dim: int, alpha: float) -> None:
        r"""Diagonal diffusion matrix network based on a constant layer.

        This implements the diffusion matrix function that is constant
        $$
            \sigma(x, args) = \text{diag}(\alpha + \text{Constant}^2)^{\frac{1}{2}}.
        $$
        where $\text{Constant}$ is a vector of size `dim`.
        Note that by constant we mean that it does not depend on the input $x$ or the parameters $args$,
        but it can be trained.

        Args:
            key (PRNGKey): random key
            dim (int): dimension of the input
            alpha (float): regulariser
        """
        self.dim = dim
        self.alpha = alpha
        self.constant_layer = ConstantLayer(dim, key)

    def __call__(self, x: ArrayLike, args: ArrayLike) -> Array:
        sigma_diag = self.constant_layer()
        sigma_squared_regularised = jnp.sqrt(self.alpha + sigma_diag**2)
        return jnp.diag(sigma_squared_regularised)


# ------------------------------------------------------------------ #
#                      Dimensionality transforms                     #
# ------------------------------------------------------------------ #


class PCATransform(eqx.Module):
    """PCA transform."""

    mean: ArrayLike
    components: ArrayLike
    scaling: ArrayLike
    centre: bool

    def __init__(
        self,
        mean: ArrayLike,
        components: ArrayLike,
        scaling: ArrayLike,
        centre: bool = False,
    ) -> None:
        r"""PCA transform.

        Transforms the input vector $x$ via
        $$
            x \mapsto \text{components} ( x ) / \sqrt{\text{scaling}}.
        $$
        If `centre` is set to `True`, then the input is first centered by subtracting the `mean`.

        Args:
            mean (ArrayLike): mean of the data used to fit the PCA
            components (ArrayLike): PCA components
            scaling (ArrayLike): scaling of the PCA transform (e.g. explained variance)
            centre (bool, optional): whether to center the data using `mean`. Defaults to False.
        """
        self.mean = jnp.array(mean)
        self.components = jnp.array(components)
        self.scaling = jnp.array(scaling)
        self.centre = centre

    def __call__(self, x: ArrayLike) -> Array:
        if self.centre:
            x = x - self.mean
        x_reduced = self.components @ x
        return x_reduced / jnp.sqrt(self.scaling)


class InversePCATransform(PCATransform):
    """Inverse PCA transform."""

    def __init__(
        self, mean: ArrayLike, components: ArrayLike, scaling: ArrayLike
    ) -> None:
        r"""Inverse PCA transform.

        Transforms the input vector $z$ via
        $$
            z \mapsto \text{components}^\top ( z \sqrt{\text{scaling}} )
        $$
        If `centre` is set to `True`, `mean` is added to the output.

        Args:
            mean (ArrayLike): mean of the data used to fit the PCA
            components (ArrayLike): PCA components
            scaling (ArrayLike): scaling of the PCA transform (e.g. explained variance)
        """
        super().__init__(mean, components, scaling)

    def __call__(self, z: ArrayLike) -> Array:
        x_reconstructed = self.components.T @ (jnp.sqrt(self.scaling) * z)
        if self.centre:
            x_reconstructed = x_reconstructed + self.mean
        return x_reconstructed


class PCAResNetTransform(PCATransform):
    mlp: MLP
    mlp_scale: float
    mlp_input_scale: float

    def __init__(
        self,
        mean: ArrayLike,
        components: ArrayLike,
        scaling: ArrayLike,
        key: PRNGKey,
        units: list[int],
        activation: str,
        mlp_scale: float,
        mlp_input_scale: float,
    ) -> None:
        r"""PCA-ResNet transform.

        This combines the PCA transform with an MLP to give a ResNet-like architecture.
        $$
            x \mapsto \text{PCA}(x) + \text{mlp\_scale} \times \text{MLP}(\text{mlp\_input\_scale} \times x).
        $$
        The input scale adjusts the input (often not order 1) so that the MLP can learn the correct scale.

        Args:
            mean (ArrayLike): mean of the data used to fit the PCA
            components (ArrayLike): PCA components
            scaling (ArrayLike): scaling of the PCA transform (e.g. explained variance)
            key (PRNGKey): random key
            units (list[int]): layer sizes of the MLP
            activation (str): activation function (can be any in `jax.nn` or custom ones defined in `onsagernet._activations`)
            mlp_scale (float): scale of the MLP output
            mlp_input_scale (float): scale of the input to the MLP
        """
        super().__init__(mean, components, scaling)
        units = units + [components.shape[0]]
        dim = components.shape[1]
        self.mlp = MLP(key, dim, units, activation)
        self.mlp_scale = mlp_scale
        self.mlp_input_scale = mlp_input_scale

    def __call__(self, x: ArrayLike) -> Array:
        pca_features = super().__call__(x)
        mlp_features = self.mlp(self.mlp_input_scale * x)
        return pca_features + self.mlp_scale * mlp_features

    def pca_transform(self, x: ArrayLike) -> Array:
        """Perform the PCA transform.

        Args:
            x (ArrayLike): state

        Returns:
            Array: pca features
        """
        return super().__call__(x)


class InversePCAResNetTransform(InversePCATransform):
    """Inverse PCA-ResNet transform."""

    mlp: MLP
    mlp_scale: float

    def __init__(
        self,
        mean: ArrayLike,
        components: ArrayLike,
        scaling: ArrayLike,
        key: PRNGKey,
        units: list[int],
        activation: str,
        mlp_scale: float,
    ) -> None:
        r"""Inverse PCA-ResNet transform.

        This combines the inverse PCA transform with an MLP to give a ResNet-like architecture.
        $$
            z \mapsto \text{PCA}^{-1}(z) + \text{mlp\_scale} \times \text{MLP}(z).
        $$

        Args:
            mean (ArrayLike): mean of the data used to fit the PCA
            components (ArrayLike): PCA components
            scaling (ArrayLike): scaling of the PCA transform (e.g. explained variance)
            key (PRNGKey): random key
            units (list[int]): layer sizes of the MLP
            activation (str): activation function (can be any in `jax.nn` or custom ones defined in `onsagernet._activations`)
            mlp_scale (float): scale of the MLP output
        """
        super().__init__(mean, components, scaling)
        units = units + [components.shape[1]]
        dim = components.shape[0]
        self.mlp = MLP(key, dim, units, activation)
        self.mlp_scale = mlp_scale

    def __call__(self, z: ArrayLike) -> Array:
        inverse_pca_recon = super().__call__(z)
        mlp_recon = self.mlp(z)
        return inverse_pca_recon + self.mlp_scale * mlp_recon

    def inverse_pca_transform(self, z: ArrayLike) -> Array:
        """Inverse PCA transform.

        Args:
            z (ArrayLike): reduced state

        Returns:
            Array: reconstructed state
        """
        return super().__call__(z)



"""
# Utilities
"""

import jax
import jax.numpy as jnp

# ------------------------- Typing imports ------------------------- #

from jax.typing import DTypeLike


def default_floating_dtype() -> DTypeLike:
    """Returns the default dtype for floating point operations.

    Returns:
        DTypeLike: default data type
    """
    if jax.config.jax_enable_x64:
        return jnp.float64
    else:
        return jnp.float32



"""
# OnsagerNet package

This package implements the basic routines for building and training
(with or without model reduction/closure modelling)
OnsagerNet and variants.

## Main Modules
- `onsagernet.dynamics`: Dynamic models, including SDEs, OnsagerNet, and variants
- `onsagernet.transformations`: Transformations for dimensionality reduction and reconstruction
- `onsagernet.trainers`: Training routines for SDEs and those with dimensionality transformations
- `onsagernet.models`: Basic model definitions (mostly fully connected neural networks) that are used in references [1-3]

## References
1. Chen, X. et al. Constructing custom thermodynamics using deep learning. Nature Computational Science 4, 66-85 (2024).
2. Novoselov, K. S. & Li, Q. Learning physical laws from observations of complex dynamics. Nature Computational Science 1-2 (2024).
3. Yu, H., Tian, X., E, W. & Li, Q. OnsagerNet: Learning stable and interpretable dynamics using a generalized Onsager principle. Phys. Rev. Fluids 6, 114402 (2021).
"""



"""
# Custom activations

This module contains custom activation functions.
"""

import jax
import jax.numpy as jnp

# ------------------------- Typing imports ------------------------- #

from jax import Array
from jax.typing import ArrayLike
from typing import Callable


# ------------------------------------------------------------------ #
#                     Custom activation functions                    #
# ------------------------------------------------------------------ #


@jax.jit
def recu(x: ArrayLike) -> Array:
    r"""Rectified cubic unit activation function.

    Implements the activation function
    $$
        x \mapsto
        \begin{cases}
            x^3 / 3 & \text{if } x \in [0, 1), \qquad \\
            x - 2/3 & \text{if } x \in [1, \infty).
        \end{cases}
    $$

    Args:
        x (ArrayLike): inputs

    Returns:
        Array: activated inputs
    """
    cubic_part = x**3 / 3
    linear_part = x - 2 / 3
    return jnp.where(x < 0, 0, jnp.where(x < 1, cubic_part, linear_part))


@jax.jit
def srequ(x: ArrayLike) -> Array:
    r"""Shifted rectified quadratic unit activation function.

    Implements the activation function
    $$
        x \mapsto
        \max(0, x)^2 - \max(0, x - 0.5)^2.
    $$

    Args:
        x (ArrayLike): inputs

    Returns:
        Array: activated inputs
    """
    return jnp.maximum(0, x) ** 2 - jnp.maximum(0, x - 0.5) ** 2


# ------------------------------------------------------------------ #
#                               Helpers                              #
# ------------------------------------------------------------------ #

CUSTOM_ACTIVATIONS = {
    "recu": recu,
    "srequ": srequ,
}


def get_activation(name: str) -> Callable[[ArrayLike], Array]:
    """Get the activation function by name.
    First checks if the activation function is a custom activation, then tries to get it from `jax.nn`.

    Args:
        name (str): name of the activation function

    Raises:
        ValueError: If the activation function is not found in custom activations or `jax.nn`
        TypeError: If the activation function is not callable

    Returns:
        _type_: activation function
    """

    # Check custom activations first
    if name in CUSTOM_ACTIVATIONS:
        activation_function = CUSTOM_ACTIVATIONS[name]
    else:
        # Try getting the activation function from jax.nn
        try:
            activation_function = getattr(jax.nn, name)
        except AttributeError:
            raise ValueError(
                f"Activation function '{name}' not found in custom activations or jax.nn"
            )

    # Check if the result is callable (i.e., a function)
    if not callable(activation_function):
        raise TypeError(f"The activation function '{name}' is not callable")

    return activation_function



r'''
# Dynamical models

This module defines the dynamical models used in the OnsagerNet and related architectures.
Note that the actual model components' architectures are not defined here, but rather it is the
assembly logic for the model components that is handled here.

The following is an example of how to use the `OnsagerNet` model
to create the stochastic OnsagerNet dynamics

$$
    dX(t) = -
    \left[
        M(X(t)) + W(x(t))
    \right] \nabla V(x(t), u(t)) dt
    + \sqrt{\epsilon} \sigma(x(t), u(t)) dW(t)
    \qquad
    X(t) \in \mathbb{R}^d, \quad u(t) \in \mathbb{R}^m.
$$

```python
import equinox as eqx
from onsagernet.dynamics import OnsagerNet

class MyPotential(eqx.Module):
    """Implement your V function here

    This should be a function (d + m) -> (1)
    """

class MyDissipation(eqx.Module):
    """Implement your M function here

    This should be a function (d) -> (d, d)
    """

class MyConservation(eqx.Module):
    """Implement your W function here

    This should be a function (d) -> (d, d)
    """

class MyDiffusion(eqx.Module):
    """Implement your sigma function here

    This should be a function (d + m) -> (d, d)
    """


potential = MyPotential()
dissipation = MyDissipation()
conservation = MyConservation()
diffusion = MyDiffusion()

sde = OnsagerNet(
    potential=potential,
    dissipation=dissipation,
    conservation=conservation,
    diffusion=diffusion,
)
```

The `sde` instance can then be used to simulate the dynamics
of the system or perform training.

- Some simple definition of the potential, dissipation, conservation, and diffusion functions are
  provided in `onsagernet.models`
- The `ReducedSDE` class includes both an `SDE` component and a dimensionality reduction component
  involving both an `onsagernet.transformations.Encoder` and a `onsagernet.transformations.Decoder`
- Standard training routines are provided in `onsagernet.trainers`
'''

import jax
import jax.numpy as jnp
import equinox as eqx
from abc import abstractmethod

# ------------------------- Typing imports ------------------------- #

from typing import Callable
from jax.typing import ArrayLike
from jax import Array
from .transformations import Encoder, Decoder

DynamicCallable = Callable[[ArrayLike, ArrayLike, ArrayLike], Array]

# ------------------------------------------------------------------ #
#                                 SDE                                #
# ------------------------------------------------------------------ #


class SDE(eqx.Module):
    """Base class for stochastic differential equations models."""

    @abstractmethod
    def drift(self, t: ArrayLike, x: ArrayLike, args: ArrayLike) -> Array:
        """Drift function

        Args:
            t (ArrayLike): time
            x (ArrayLike): state
            args (ArrayLike): additional arguments or parameters

        Returns:
            Array: drift vector field
        """
        pass

    @abstractmethod
    def diffusion(self, t: ArrayLike, x: ArrayLike, args: ArrayLike) -> Array:
        """Diffusion function

        Args:
            t (ArrayLike): time
            x (ArrayLike): state
            args (ArrayLike): additional arguments or parameters

        Returns:
            Array: diffusion matrix of size (state_dim, bm_dim)
        """
        pass


class SDEfromFunc(SDE):
    """SDE model defined by providing drift and diffusion functions."""

    _drift_func: DynamicCallable
    _diffusion_func: DynamicCallable

    def __init__(
        self, drift_func: DynamicCallable, diffusion_func: DynamicCallable
    ) -> None:
        r"""SDE model defined by providing drift and diffusion functions.

        This implements the dynamics
        $$
            dX(t) = f(t, X(t), u(t)) dt + g(t, X(t), u(t)) dW(t)
        $$
        where $f$ is the drift function and $g$ is the diffusion function.
        The `args` argument (which represents $u(t)$)
        is used to pass additional parameters to the drift and diffusion functions.

        Args:
            drift_func (Callable[[ArrayLike, ArrayLike, ArrayLike], Array]): provided drift function
            diffusion_func (Callable[[ArrayLike, ArrayLike, ArrayLike], Array]): provided diffusion function
        """

        self._drift_func = drift_func
        self._diffusion_func = diffusion_func

    def drift(self, t: ArrayLike, x: ArrayLike, args: ArrayLike) -> Array:
        """Drift function

        Args:
            t (ArrayLike): time
            x (ArrayLike): state
            args (ArrayLike): additional arguments or parameters

        Returns:
            Array: drift vector field
        """
        return self._drift_func(t, x, args)

    def diffusion(self, t: ArrayLike, x: ArrayLike, args: ArrayLike) -> Array:
        """Diffusion function

        Args:
            t (ArrayLike): time
            x (ArrayLike): state
            args (ArrayLike): addional arguments or parameters

        Returns:
            Array: diffusion matrix
        """
        return self._diffusion_func(t, x, args)


class ReducedSDE(eqx.Module):
    """SDE model with encoder and decoder with dimensionality reduction or closure modelling."""

    encoder: Encoder
    decoder: Decoder
    sde: SDE

    def __init__(self, encoder: Encoder, decoder: Decoder, sde: SDE) -> None:
        """SDE model with encoder and decoder with dimensionality reduction or closure modelling.

        The `sde` attribute can be any model of the [SDE](#SDE) class or its sub-classes.
        The `encoder` must be a (sub-)class of `onsagernet.transformations.Encoder` and
        the `decoder` must be a (sub-)class of `onsagernet.transformations.Decoder`.

        Args:
            encoder (Encoder): The encoder function mapping the microscopic state to the reduced state
            decoder (Decoder): The decoder function mapping the reduced state to the microscopic state
            sde (SDE): The stochastic dynamics for the reduced state
        """
        self.encoder = encoder
        self.decoder = decoder
        self.sde = sde


# ------------------------------------------------------------------ #
#                             OnsagerNet                             #
# ------------------------------------------------------------------ #


class OnsagerNet(SDE):
    potential: eqx.Module
    dissipation: eqx.Module
    conservation: eqx.Module
    diffusion_func: eqx.Module

    def __init__(
        self,
        potential: eqx.Module,
        dissipation: eqx.Module,
        conservation: eqx.Module,
        diffusion: eqx.Module,
    ) -> None:
        r"""Stochastic OnsagerNet model.

        Let $X(t) \in \mathbb{R}^d$. The Stochastic OnsagerNet model is defined by the SDE
        $$
            dX(t) = -
            \left[
                M(X(t)) + W(x(t))
            \right] \nabla V(x(t), u(t)) dt
            + \sqrt{\epsilon} \sigma(x(t), u(t)) dW(t)
        $$
        where

        - $M : \mathbb{R}^{d} \to \mathbb{R}^{d\times d}$ is the dissipation matrix, which is symmetric positive semi-definite for all $x$
        - $W : \mathbb{R}^{d} \to \mathbb{R}^{d\times d}$ is the conservation matrix, which is anti-symmetric for all $x$
        - $V : \mathbb{R}^{d} \to \mathbb{R}$ is the potential function
        - $\sigma: \mathbb{R}^{d} \to \mathbb{R}^{d\times d}$ is the (square) diffusion matrix
        - $u(t)$ are the additional parameters for the potential and diffusion functions, and note that **the first dimension of $u(t)$ is the temperature $\epsilon$**

        Args:
            potential (eqx.Module): potential function $V$
            dissipation (eqx.Module): dissipation matrix $M$
            conservation (eqx.Module): conservation matrix $W$
            diffusion (eqx.Module): diffusion matrix $\sigma$
        """
        self.potential = potential
        self.dissipation = dissipation
        self.conservation = conservation
        self.diffusion_func = diffusion

    def drift(self, t: ArrayLike, x: ArrayLike, args: ArrayLike) -> Array:
        """Drift function

        Args:
            t (ArrayLike): time
            x (ArrayLike): state
            args (ArrayLike): additional arguments or parameters, the first element is the temperature

        Returns:
            Array: drift vector field
        """
        dvdx = jax.grad(self.potential, argnums=0)(x, args)
        return -(self.dissipation(x) + self.conservation(x)) @ dvdx

    def diffusion(self, t: ArrayLike, x: ArrayLike, args: ArrayLike) -> Array:
        """Diffusion function

        Args:
            t (ArrayLike): time
            x (ArrayLike): state
            args (ArrayLike): additional arguments or parameters, the first element is the temperature

        Returns:
            Array: diffusion matrix
        """
        temperature = args[0]
        return jnp.sqrt(temperature) * self.diffusion_func(x, args)


# ------------------------------------------------------------------ #
#        OnsagerNet satifying fluctuation-dissipation relation       #
# ------------------------------------------------------------------ #


class OnsagerNetFD(OnsagerNet):
    shared: eqx.nn.Shared

    def __init__(
        self, potential: eqx.Module, dissipation: eqx.Module, conservation: eqx.Module
    ) -> None:
        r"""Stochastic OnsagerNet model satisfying a fluctuation-dissipation relation.

        This is a modified version of the Stochastic OnsagerNet model.
        Let $X(t) \in \mathbb{R}^d$. This model is defined by the SDE
        $$
            dX(t) = -
            \left[
                M(X(t)) + W(x(t))
            \right] \nabla V(x(t), u(t)) dt
            + \sqrt{2 \epsilon} [M(x(t)]^\frac{1}{2}dW(t)
        $$
        where

        - $M : \mathbb{R}^{d} \to \mathbb{R}^{d\times d}$ is the dissipation matrix,
          which is symmetric positive semi-definite for all $x$
        - $W : \mathbb{R}^{d} \to \mathbb{R}^{d\times d}$ is the conservation matrix,
          which is anti-symmetric for all $x$
        - $V : \mathbb{R}^{d} \to \mathbb{R}$ is the potential function
        - $u(t)$ are the additional parameters for the potential and diffusion functions,
          and note that **the first dimension of $u(t)$ is the temperature $\epsilon$**

        Notice that the main difference with `OnsagerNet` is that the
        diffusion matrix is now given by a (positive semi-definite) square root of the dissipation matrix.

        Args:
            potential (eqx.Module): potential function $V$
            dissipation (eqx.Module): dissipation matrix $M$
            conservation (eqx.Module): conservation matrix $W$
        """
        self.potential = potential
        self.conservation = conservation
        self.diffusion_func = None

        # Share the dissipation module
        dissipation_drift = dissipation
        dissipation_diffusion = dissipation
        where = lambda shared_layers: shared_layers[0]
        get = lambda shared_layers: shared_layers[1]
        self.shared = eqx.nn.Shared(
            (dissipation_drift, dissipation_diffusion), where, get
        )

    @property
    def dissipation(self) -> eqx.Module:
        """Dissipation matrix wrapper

        Returns:
            eqx.Module: dissipation matrix module
        """
        return self.shared()[0]

    def _matrix_div(self, M: eqx.Module, x: ArrayLike) -> Array:
        r"""Computes the matrix divergence of a matrix function $M(x)$.

        This is defined in component form as
        $$
            [\nabla \cdot M(x)]_i = \sum_j \frac{\partial M_{ij}}{\partial x_j}.
        $$

        Args:
            M (eqx.Module): matrix function
            x (ArrayLike): state

        Returns:
            Array: \nabla \cdot M(x)
        """
        jac_M_x = jax.jacfwd(M)(x)
        return jnp.trace(jac_M_x, axis1=1, axis2=2)

    def drift(self, t: ArrayLike, x: ArrayLike, args: ArrayLike) -> Array:
        """Drift function

        Args:
            t (ArrayLike): time
            x (ArrayLike): state
            args (ArrayLike): additional arguments or parameters, the first element is the temperature

        Returns:
            Array: drift vector field
        """
        temperature = args[0]
        dissipation = self.shared()[0]
        f1 = super().drift(t, x, args)
        f2 = temperature * self._matrix_div(dissipation, x)
        return f1 + f2

    def diffusion(self, t: ArrayLike, x: ArrayLike, args: ArrayLike) -> Array:
        """Diffusion function

        Args:
            t (ArrayLike): time
            x (ArrayLike): state
            args (ArrayLike): additional arguments or parameters, the first element is the temperature

        Returns:
            Array: diffusion matrix
        """
        temperature = args[0]
        dissipation = self.shared()[1]
        M_x = dissipation(x)
        sqrt_M_x = jnp.linalg.cholesky(M_x)
        return jnp.sqrt(2.0 * temperature) * sqrt_M_x



r"""
# Basic trainers for SDE models

This module implements basic training routines for SDE models.
The base class is `SDETrainer`, which provides the base training logic.
The sub-classes are required to implement the `SDETrainer.loss_func` that is used to train the model.

## Training routines

We provide here two training routines.

- `MLETrainer`: this implements the maximum likelihood loss [`MLELoss`](./_losses.html#MLELoss)
  to estimate SDE drift and diffusion
  through computing the [maximum-likelihood](https://en.wikipedia.org/wiki/Maximum_likelihood_estimation) following the
  [Euler-Maruyama discretisation](https://en.wikipedia.org/wiki/Euler–Maruyama_method) of the SDE.
- `ClosureMLETrainer`: this combines [`MLELoss`](./_losses.html#MLELoss) with additional losses to enforce closure,
  namely a reconstruction loss [`ReconLoss`](./_losses.html#ReconLoss) and a
  comparison loss [`CompareLoss`](./_losses.html#CompareLoss).
  This follows the implementation of [1].

The following example shows how to train an `onsagernet.dynamics.OnsagerNet` model using the `MLETrainer`.
```python
from onsagernet.dynamics import OnsagerNet
from onsagernet.trainers import MLETrainer

sde = OnsagerNet(...)
dataset = load_data(...)  # return a datasets.Dataset object

trainer = MLETrainer(opt_options=config.train.opt, rop_options=config.train.rop)
sde, losses, _ = trainer.train(  # trains the model `sde` for 10 epochs with batch size 2
    model=sde,
    dataset=dataset,
    num_epochs=10,
    batch_size=2,  # batch size should be typically small since this yields [n_batch, n_steps, n_dim] data
)
```

## Dataset format
The dataset is assumed
to be a huggingface [`datasets.Dataset`](https://huggingface.co/docs/datasets/en/index)
object with three columns: `t`, `x`, and `args`.
Here, `t` is time, `x` is state, and `args` are additional arguments.
Let `n_examples` be the number of examples, `n_dim` be the dimension of the state x,
`n_args` be the dimension of the arguments, and `n_steps` be the number of time steps.
Then, the shapes of the data entries are as follows:
- t: (`n_examples`, `n_steps`, `1`) - this is the sequence of time steps sampled in the time series,
  which can vary from one example to another (but the `n_steps` must be the same), so consider padding
  if this were not the case
- x: (`n_examples`, `n_steps`, `n_dim`) - this is the corresponding sequence of states
- args: (`n_examples`, `n_steps`, `n_args`) - this is the sequence of additional arguments,
  which can be constant or varying in time. If constant, you must broad-cast the values
  along the time dimension so that the expected shapes are maintained.

We also assume that `n_args` >= `1`, with the first dimension always being
temperature, since it is relevant for all SDE models in physical modelling.
Your custom `loss_func` routine must return a scalar loss value.

*In fact, it is not strictly necessary to use a [`datasets.Dataset`](https://huggingface.co/docs/datasets/en/index)
object, but any object that can yield the correct data format by calling `dataset.iter(batch_size)`.*

## Writing custom training routines

You can write custom training routines (with custom losses, for example)
by sub-classing `SDETrainer`.
Sub-classes must implement the `SDETrainer.loss_func` method.

Given a model partitioned by `eqx.partition(model, ...)`,
into a trainable part `diff_model` and a static part `static_model`,
the `SDETrainer.loss_func` routine computes the loss of the model on the given dataset.
We always assume that the batch data is a tuple of `(t, x, args)`.

See above for more on the format of the dataset and
`MLETrainer.loss_func`, or the more complex `ClosureMLETrainer.loss_func` for examples.

## References

1. Chen, X. et al. *Constructing custom thermodynamics using deep learning*. Nature Computational Science **4**, 66–85 (2024).


"""

import os
import jax
import jax.numpy as jnp
import equinox as eqx
from abc import ABC, abstractmethod

from optax import adam, chain
from optax.contrib import reduce_on_plateau
from optax.tree_utils import tree_get
from jax.tree_util import tree_map

from tqdm import tqdm

from ._losses import MLELoss, ReconLoss, CompareLoss

# ------------------------- Typing imports ------------------------- #

from chex import ArrayTree  # to silence pdoc warnings
from .dynamics import SDE, ReducedSDE
from typing import Optional, Any, Union
from jax.typing import ArrayLike
from jax import Array
from datasets import Dataset
from optax import GradientTransformation, OptState
from logging import Logger

DynamicModel = Union[SDE, ReducedSDE]

# ------------------------------------------------------------------ #
#                              Trainers                              #
# ------------------------------------------------------------------ #


class SDETrainer(ABC):
    """Base class for training SDE models."""

    def __init__(
        self, opt_options: dict, rop_options: dict, loss_options: Optional[dict] = None
    ) -> None:
        """SDE training routine.

        Args:
            opt_options (dict): dictionary of options for the optimiser
            rop_options (dict): dictionary of options for the reduce-on-plateau callback
            loss_options (Optional[dict], optional): dictionary of options for loss computation. Defaults to None.
        """
        self._opt_options = opt_options
        self._rop_options = rop_options
        self._loss_options = loss_options

    @eqx.filter_jit
    @abstractmethod
    def loss_func(
        self,
        diff_model: DynamicModel,
        static_model: DynamicModel,
        t: ArrayLike,
        x: ArrayLike,
        args: ArrayLike,
    ) -> float:
        """Loss function.

        This must be implemented by sub-classes.

        Args:
            diff_model (DynamicModel): the trainable part of the model
            static_model (DynamicModel): the static part of the model
            t (ArrayLike): time
            x (ArrayLike): state
            args (ArrayLike): additional arguments or parameters. The first dimension is temperature.

        Returns:
            float: loss value
        """
        pass

    def _make_optimiser(
        self, opt_options: dict, rop_options: dict
    ) -> GradientTransformation:
        """Make an optimiser.

        Args:
            opt_options (dict): optimiser options
            rop_options (dict): reduce-on-plateau options

        Returns:
            GradientTransformation: an optimiser object from `optax`
        """
        return chain(adam(**opt_options), reduce_on_plateau(**rop_options))

    @eqx.filter_jit
    def _make_step(
        self,
        model: DynamicModel,
        data: Dataset,
        opt: GradientTransformation,
        opt_state: OptState,
        filter_spec: Any,
    ) -> tuple[DynamicModel, OptState, float]:
        """Make a training step.

        Args:
            model (DynamicModel): the model to be trained
            data (Dataset): the dataset object
            opt (GradientTransformation): optimiser object
            opt_state (OptState): optimiser state
            filter_spec (Any): the filtering logic to determine which parts of the model to train

        Returns:
            tuple[DynamicModel, OptState, float]: trained model, optimiser state, loss value
        """
        diff_model, static_model = eqx.partition(model, filter_spec)

        loss_value, grads = eqx.filter_value_and_grad(self.loss_func)(
            diff_model, static_model, *data
        )
        updates, opt_state = opt.update(grads, opt_state, model, value=loss_value)
        model = eqx.apply_updates(model, updates)
        return model, loss_value, opt_state

    def _train_epoch(
        self,
        model: DynamicModel,
        dataset: Dataset,
        batch_size: int,
        opt: GradientTransformation,
        opt_state: OptState,
        filter_spec: Any,
    ) -> tuple[DynamicModel, float, OptState]:
        """Train the model for an epoch.

        Args:
            model (DynamicModel): the model to be trained
            dataset (Dataset): the dataset object
            batch_size (int): the batch size
            opt (GradientTransformation): the optimiser object
            opt_state (OptState): the optimiser state
            filter_spec (Any): the filtering logic to determine which parts of the model to train

        Returns:
            tuple[DynamicModel, float, OptState]: trained model, loss value, optimiser state
        """

        step_losses = []

        for batch in tqdm(
            dataset.iter(batch_size),
            total=dataset.num_rows // batch_size,
        ):
            data_batch = (batch["t"], batch["x"], batch["args"])
            model, train_loss, opt_state = self._make_step(
                model, data_batch, opt, opt_state, filter_spec
            )
            step_losses.append(train_loss)
        epoch_loss = jnp.mean(jnp.array(step_losses))
        return model, epoch_loss, opt_state

    def train(
        self,
        model: DynamicModel,
        dataset: Dataset,
        num_epochs: int,
        batch_size: int,
        logger: Optional[Logger] = None,
        opt_state: Optional[OptState] = None,
        filter_spec: Optional[Any] = None,
        checkpoint_dir: Optional[str] = None,
        checkpoint_every: Optional[int] = None,
    ) -> tuple[DynamicModel, list[float], OptState]:
        """The main training routine.

        Args:
            model (DynamicModel): the model to be trained
            dataset (Dataset): the dataset
            num_epochs (int): number of epochs to train
            batch_size (int): the batch size
            logger (Optional[Logger], optional): the logging object. Defaults to None.
            opt_state (Optional[OptState], optional): the starting optimiser state. Defaults to None.
            filter_spec (Optional[Any], optional): the filtering logic. Defaults to None.
            checkpoint_dir (Optional[str], optional): the directory to save checkpoints. Defaults to None.
            checkpoint_every (Optional[int], optional): checkpoints are saved every `checkpoint_every` number of epochs. Defaults to None.

        Returns:
            tuple[DynamicModel, list[float], OptState]: trained model, list of losses, optimiser state
        """
        opt = self._make_optimiser(self._opt_options, self._rop_options)
        if opt_state is None:
            opt_state = opt.init(eqx.filter(model, eqx.is_array))

        if filter_spec is None:
            filter_spec = tree_map(lambda _: True, model)

        losses = []
        for epoch in range(num_epochs):
            model, step_loss, opt_state = self._train_epoch(
                model=model,
                dataset=dataset,
                batch_size=batch_size,
                opt=opt,
                opt_state=opt_state,
                filter_spec=filter_spec,
            )
            losses.append(step_loss)
            if logger:
                lr_scale = tree_get(opt_state, "scale")
                logger.info(
                    f"epoch={epoch:05d}, loss={step_loss:.6f}, lr_scale={lr_scale:.4f}"
                )

            if checkpoint_dir is not None and epoch % checkpoint_every == 0:
                model_path = os.path.join(
                    checkpoint_dir, f"model_epoch_{epoch:05d}.eqx"
                )
                eqx.tree_serialise_leaves(model_path, model)

        return model, losses, opt_state


class MLETrainer(SDETrainer):

    @eqx.filter_jit
    def loss_func(
        self,
        diff_model: DynamicModel,
        static_model: DynamicModel,
        t: ArrayLike,
        x: ArrayLike,
        args: ArrayLike,
    ) -> float:
        """The MLE loss function.

        See [`MLELoss`](./_losses.html#MLELoss) for more details.

        Args:
            diff_model (DynamicModel): the trainable part of the model
            static_model (DynamicModel): the static part of the model
            t (ArrayLike): time
            x (ArrayLike): state
            args (ArrayLike): additional arguments or parameters.

        Returns:
            float: the computed loss
        """
        model = eqx.combine(diff_model, static_model)
        return MLELoss()(model, t, x, args)


class ClosureMLETrainer(MLETrainer):

    @eqx.filter_jit
    def loss_func(
        self,
        diff_model: DynamicModel,
        static_model: DynamicModel,
        t: ArrayLike,
        x: ArrayLike,
        args: ArrayLike,
    ) -> float:
        r"""The combined loss function for MLE training with closure modelling.

        The losses are applied to the combine model
        `model = eqx.combine(diff_model, static_model)`
        with three parts
        - [`MLELoss`](./_losses.html#MLELoss) applied to `model.sde`
        - [`ReconLoss`](./_losses.html#ReconLoss) applied to the `model`
        - [`CompareLoss`](./_losses.html#CompareLoss) applied to the `model`

        The combined loss is given by
        $$
            \text{loss} = \text{loss_sde}
            + \text{recon_weight} \times \text{loss_recon}
            + \text{compare_weight} \times \text{loss_compare}
        $$

        The variables `recon_weight` and `compare_weight` are set
        in the `loss_options` attribute.

        Args:
            diff_model (DynamicModel): the trainable part of the model
            static_model (DynamicModel): the static part of the model
            t (ArrayLike): time
            x (ArrayLike): state
            args (ArrayLike): additional arguments or parameters.

        Returns:
            float: the computed loss
        """
        model = eqx.combine(diff_model, static_model)
        z = jax.vmap(jax.vmap(model.encoder))(x)
        loss_sde = MLELoss()(model.sde, t, z, args)
        loss_recon = ReconLoss()(model, x)
        loss_compare = CompareLoss()(model, x)
        return (
            loss_sde
            + self._loss_options["recon_weight"] * loss_recon
            + self._loss_options["compare_weight"] * loss_compare
        )



"""Custom layers as equinox modules"""

import jax
import equinox as eqx
from math import sqrt
from ._utils import default_floating_dtype
from jax import Array
from jax.random import PRNGKey
from typing import Optional
from jax.typing import DTypeLike


class ConstantLayer(eqx.Module):
    """Constant layer."""

    weight: Array
    dim: int

    def __init__(
        self, dim: int, key: PRNGKey, dtype: Optional[DTypeLike] = None
    ) -> None:
        """Constant layer.

        Returns a constant, trainable vector,
        similar to the bias term in a neural network,
        that is the same size as the input.

        Args:
            dim (int): dimension of the input space
            key (PRNGKey): random key
            dtype (Optional[DTypeLike], optional): data type. Defaults to None.
        """
        dtype = default_floating_dtype() if dtype is None else dtype
        lim = 1 / sqrt(dim)
        shape = (dim,)
        self.dim = dim
        self.weight = jax.random.uniform(
            key, shape, minval=-lim, maxval=lim, dtype=dtype
        )

    def __call__(self, *, key: Optional[PRNGKey] = None) -> Array:
        return self.weight
